#!/usr/bin/env python

# pbm2popeye 0.1.0 Copyright 2008 
# Author: Dmitri Turevski <dmitri.turevski@gmail.com>
# 
# This program is distributed under the terms of the GNU General Public License
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""NAME
    pbm2popeye - convert Problemiste PBM files to Popeye input format.

SYNOPSIS
    pbm2popeye.py file

DESCRIPTION
    The result is written to the standart output. BeginProblem, EndProblem and
    trailing NextProblem commands are not included to allow easier gluing of
    multiple outputs. Twins types are supported partially.
    
BUGS
    If you find a bug, please report it to dmitri.turevski@gmail.com

AUTHOR
    Dmitri Turevski <dmitri.turevski@gmail.com>
"""

import sys, array

def main():
    if len(sys.argv) != 2:
        print __doc__
        sys.exit(1)

    try:
        input = open(sys.argv[1])
    except IOError:
        sys.stderr.write("Can't open '" + sys.argv[1] + "' for reading\n")
        sys.exit(2)
    
    try:
        num_entries = read_int(input, 'H') + 1
        for i in xrange(num_entries):
            try:
                r = Record(input)
                r.comments.append('file://' + sys.argv[1] + '/' + str(i))
                r.dump(i != num_entries - 1)
            except:
                sys.stderr.write("File corrupted: " + sys.argv[1] + \
                    ' #' + str(i) +"\n")
    #except IOError, IndexError:
    except:
        sys.stderr.write("File corrupted: " + sys.argv[1] + "\n")
        input.close()
        sys.exit(2)
        
    input.close()

def read_int(file, decl):
    v = array.array(decl)
    v.fromfile(file, 1)
    return v[0]

def read_string(file):
    offset = read_int(file, 'L')
    cur_pos = file.tell()
    file.seek(offset)
    string = file.read(read_int(file, 'H'))
    file.seek(cur_pos)
    return string

def byte2piece(byte):
    if byte == 32:
        return ''
    byte = byte - 33
    if (byte < 0) or (byte > 14):
        return '?'
    return 'RQBSPK???kpsbqr'[byte]

def byte2square(byte):
    byte = byte - 32
    return 'abcdefgh'[byte % 8] + '87654321'[byte // 8]

def bytes2stipulation(byte1, byte2):
    #binary = lambda i, c = (lambda i, c: i and (c(i >> 1, c) + str(i & 1)) or ''): c(i, c)
    #print binary(byte1), byte2
    
    type = ['', 'h', 's', 'r'][byte1 % 4]
    type = type + ['#', '='][(byte1 & (1 << 4)) != 0]
    if byte1 & (1 << 3) :
        type = 'Ser-' + type
    num_moves = byte2
    if byte1 & (1 << 6) :
        num_moves = str(num_moves - 1) + '.5'
    
    return type + str(num_moves), byte1 & (1 << 5)

class Record:
    def __init__(self, file):
        read_int(file, 'H') # skip id
        self.stipulation, self.is_maximummer = \
            bytes2stipulation(read_int(file, 'B'), read_int(file, 'B'))
        
        self.info = {}
        for field in ['author', 'source', 'distinction', 'comments', 'text', \
            'twins', 'extra']:
            self.info[field] = read_string(file)
            
        self.comments = []
        for field in ['comments', 'text', 'extra']:
            if self.info[field]:
                for line in self.info[field].split("\n"):
                    self.comments.append(field + ': ' + line.strip())
        
        self.parse_twins()
        
        self.fen, blanks = '', 0
        for i in xrange(64):
            if((i > 0) and (i % 8 == 0)): # new row
                self.fen = self.fen + ['', str(blanks)][blanks > 0]
                self.fen = self.fen + "/"
                blanks = 0
            piece = byte2piece(read_int(file, 'B'))
            if(piece != ''):
                self.fen = self.fen + ['', str(blanks)][blanks > 0]
                self.fen = self.fen + piece
                blanks = 0
            else:
                blanks = blanks + 1
        self.fen = self.fen + ['', str(blanks)][blanks > 0]

    def parse_twins(self):
        self.is_duplex = self.succesive_twining = self.is_zero = False
        self.twins = []
        if self.info['twins'] == '':
            return
        self.succesive_twining = ord(self.info['twins'][0]) == 126
        self.is_zero = True

        twins = []
        for i in xrange(len(self.info['twins']) // 4):
            #for j in xrange(4):
            #    print ord(self.info['twins'][1 + 4*i + j]) - 32,
            #print
            combined = (ord(self.info['twins'][1 + 4*i]) & (1<<6)) != 0

            type, square1, square2, piece  = \
                ord(self.info['twins'][1 + 4*i]) & 31, \
                byte2square(ord(self.info['twins'][1 + 4*i + 1])), \
                byte2square(ord(self.info['twins'][1 + 4*i + 2])), \
                byte2piece(ord(self.info['twins'][1 + 4*i + 3]))
            piece = ['white', 'black'][piece.lower() == piece] + ' ' + piece
            #print type, square1, square2, piece
            if 0 == type:
                self.is_zero = False
            elif 1 == type:
                twins.append('move %s %s' % (square1, square2))
            elif 2 == type:
                twins.append('remove %s' % (square1))
            elif 3 == type:
                twins.append('add %s%s' % (piece, square1))
            elif 4 == type:
                twins.append('remove %s add %s%s' % \
                    (square1, piece, square2))
            elif 5 == type:
                twins.append('exchange %s %s' % (square1, square2))
            elif 6 == type:
                twins.append('rotate 90')
            elif 7 == type:
                twins.append('rotate 180')
            elif 8 == type:
                twins.append('rotate 270')
            elif 9 == type:
                twins.append('shift %s %s' % (square1, square2))
            elif 10 == type:
                self.comments.append('unsupported twin: torus shift %s %s' %\
                    (square1, square2))
            elif 11 == type:
                twins.append('mirror a1<-->a8')
            elif 12 == type:
                twins.append('mirror a1<-->h1')
            elif 13 == type:
                stipulation, is_maximummer = bytes2stipulation(\
                        ord(self.info['twins'][1 + 4*i + 1]) - 32, \
                        ord(self.info['twins'][1 + 4*i + 2]) - 32 )
                twin = []
                if stipulation != self.stipulation:
                    twin.append('stipulation ' + stipulation)
                if is_maximummer and not self.is_maximummer:
                    twin.append('condition maximummer')
                if not is_maximummer and self.is_maximummer:
                    twin.append('condition ')
                twins.append(' '.join(twin))
            elif 14 == type:
                self.comments.append('unsupported twin: after the key')
            elif 15 == type:
                twins.append('PolishType')
            elif 16 == type:
                self.is_duplex = True
            
            if not combined and twins:
                self.twins.append(' '.join(twins))
                twins = []
            
    def dump(self, with_separator):
        print
        if self.info['author']:
            print "Author " + self.info['author']
        if self.info['source']:
            print "Origin " + self.info['source']
        if self.info['distinction']:
            print "Title " + self.info['distinction']
        for comment in self.comments:
            print 'Remark', comment
        if self.is_maximummer:
            print "Condition maximummer"
        print "Stipulation " + self.stipulation
        print "Forsyth " + self.fen
        if self.is_duplex:
            print "Option Duplex"
        for index, twin in enumerate(self.twins):
            print ["Twin", "ZeroPosition"][index == 0 and self.is_zero],
            print ["", "continued "][self.succesive_twining] + twin
        print
        if with_separator:
            print "NextProblem"
        
if __name__ == "__main__":
    main()